/*
 * Copy me if you can.
 * by 20h
 */

#include <u.h>
#include <libc.h>
#include <thread.h>
#include <fcall.h>
#include <ip.h>
/*
 * Copy me if you can.
 * by 20h
 */

#include <9p.h>
#include <mp.h>
#include <libsec.h>

enum
{
	Blocksize = 1024,

	ReqMagic = 	0x25609513,
	RespMagic = 0x67446698,

	CliSrvMagic = 0x0000420281861253,

	Cread = 	0x00,
	Cwrite = 	0x01,
	Cdisc = 	0x02,
};

char *host;
char *user;

uvlong size;
int debug;
int netfd;
ulong time0;

void
usage(void)
{
	fprint(2, "usage: %s [-Dd] [-m mtpt] [-s srvname] host\n", argv0);
	exits("usage");
}

enum
{
	Qroot,
	Qctl,
	Qdata,
};

#define PATH(type, n)		((type)|((n)<<8))
#define TYPE(path)			((int)(path) & 0xFF)
#define NUM(path)			((uint)(path)>>8)

typedef struct nbdreq nbdreq;
struct nbdreq {
	char magic[4];
	char type[4];
	char handle[8];
	char from[8];
	char len[4];
};

typedef struct nbdresp nbdresp;
struct nbdresp {
	char magic[4];
	char error[4];
	char handle[8];
};

int
dialnbd(void)
{
	int netfd;
	char buf[128];

	netfd = dial(netmkaddr(host, "net", "1100"), 0, 0, 0);
	if(netfd < 0)
		sysfatal("dial: %r");

	if(readn(netfd, buf, 8) != 8)
		sysfatal("read magic: %r");
	if(strncmp(buf, "NBDMAGIC", 8) != 0)
		sysfatal("magic wrong");

	if(readn(netfd, buf, 8) != 8)
		sysfatal("read clisrv: %r");

	if(readn(netfd, buf, 8) != 8)
		sysfatal("read size: %r");
	size = nhgetv(buf);

	if(read(netfd, buf, 128) < 0)
		sysfatal("read garbish: %r");

	return netfd;
}

typedef struct Tab Tab;
struct Tab
{
	char *name;
	ulong mode;
};

Tab tab[] =
{
	".",		DMDIR|0555,
	"ctl",		0666,
	"data",		0666,
};

static void
fillstat(Dir *d, uvlong path)
{
	Tab *t;

	memset(d, 0, sizeof(*d));
	d->uid = estrdup9p(user);
	d->gid = estrdup9p(user);
	d->qid.path = path;
	d->atime = d->mtime = time0;
	t = &tab[TYPE(path)];
	d->name = estrdup9p(t->name);
	d->length = size;
	d->qid.type = t->mode>>24;
	d->mode = t->mode;
}

static void
fsattach(Req *r)
{
	if(r->ifcall.aname && r->ifcall.aname[0]){
		respond(r, "invalid attach specifier");
		return;
	}
	r->fid->qid.path = PATH(Qroot, 0);
	r->fid->qid.type = QTDIR;
	r->fid->qid.vers = 0;
	r->ofcall.qid = r->fid->qid;
	respond(r, nil);
}

static void
fsstat(Req *r)
{
	fillstat(&r->d, r->fid->qid.path);
	respond(r, nil);
}

static int
rootgen(int i, Dir *d, void*)
{
	i += Qroot + 1;

	if(i <= Qdata){
		fillstat(d, i);
		return 0;
	}
	return -1;
}

static char*
fswalk1(Fid *fid, char *name, Qid *qid)
{
	int i;
	ulong path;

	path = fid->qid.path;
	if(!(fid->qid.type & QTDIR))
		return "walk in non-directory";

	if(strcmp(name, "..") == 0){
		switch(TYPE(path)){
		case Qroot:
			return nil;
		default:
			return "bug in fswalk1";
		}
	}

	i = TYPE(path) + 1;
	while(i < nelem(tab)){
		if(strcmp(name, tab[i].name) == 0){
			qid->path = PATH(i, NUM(path));
			qid->type = tab[i].mode>>24;
			return nil;
		}
		if(tab[i].mode & DMDIR)
			break;
		i++;
	}
	return "directory entry not found";
}

static void
fsopen(Req *r)
{
	respond(r, nil);
}

void
donbdreq(Req *r, uint cmd)
{
	nbdreq req;
	nbdresp resp;

	if(r->ifcall.offset > size) {
		respond(r, "block device is smaller");
		return;
	}

	if(r->ifcall.count + r->ifcall.offset >= size)
		r->ifcall.count = size - r->ifcall.offset;
	if(r->ifcall.count == 0) {
		respond(r, nil);
		return;
	}

	hnputl(req.magic, ReqMagic);
	hnputl(req.type, cmd);
	hnputv(req.from, r->ifcall.offset);
	hnputl(req.len, r->ifcall.count);

	if(write(netfd, &req, 28) != 28)
		sysfatal("write req: %r");

	if(cmd == Cwrite) {
		if(write(netfd, r->ifcall.data, r->ifcall.count) != r->ifcall.count)
			sysfatal("write data: %r");
		r->ofcall.count = r->ifcall.count;
	}

	if(readn(netfd, &resp, 16) != 16)
		sysfatal("read resp: %r");

	if(nhgetl(resp.error)) {
		respond(r, "an error occured");
		return;
	}

	if(cmd == Cread) {
		if(readn(netfd, r->ofcall.data, r->ifcall.count) != r->ifcall.count)
			sysfatal("read data: %r");
		r->ofcall.count = r->ifcall.count;
	}

	respond(r, nil);
}

static void
fsread(Req *r)
{
	char e[ERRMAX], buf[8192];
	ulong path;
	int n;

	path = r->fid->qid.path;
	switch(TYPE(path)){
	case Qroot:
		dirread9p(r, rootgen, nil);
		respond(r, nil);
		break;
	case Qctl:
		n = snprint(buf, sizeof(buf), "inquiry NDB-DISK geometry %lld %d\n", size / Blocksize, Blocksize);
		readbuf(r, buf, n);
		respond(r, nil);
		break;
	case Qdata:
		donbdreq(r, Cread);
		break;
	default:
		snprint(e, sizeof(e), "bug in fsread path=%lux", path);
		respond(r, e);
		break;
	}
}

static void
fswrite(Req *r)
{
	char e[ERRMAX];
	ulong path;

	path = r->fid->qid.path;
	switch(TYPE(path)){
	case Qctl:
		respond(r, nil);
		break;
	case Qdata:
		donbdreq(r, Cwrite);
		break;
	default:
		snprint(e, sizeof(e), "bug in fswrite path=%lux", path);
		respond(r, e);
		break;
	}
}

Srv fs = 
{
.attach=	fsattach,
.walk1=		fswalk1,
.open=		fsopen,
.read=		fsread,
.write=		fswrite,
.stat=		fsstat,
};

void
threadmain(int argc, char **argv)
{
	char *mtpt, *srvname;

	mtpt = "/n/nbd";
	srvname = nil;

	ARGBEGIN {
	case 'D':
		chatty9p++;
		break;
	case 'd':
		debug++;
		break;
	case 's':
		srvname = EARGF(usage());
		break;
	case 'm':
		mtpt = EARGF(usage());
		break;
	default:
		usage();
	} ARGEND;

	if(argc < 1)
		usage();

	time0 = time(0);
	host = argv[0];

	user = getuser();
	netfd = dialnbd();
	if(size < 0)
		sysfatal("getfilesize: %r");

	threadpostmountsrv(&fs, srvname, mtpt, MBEFORE);
	threadexits(0);
}
